/*****************************************************************************
 * $Id: JavaCA.java,v 1.7 2002/10/07 03:20:17 jheiss Exp $
 *****************************************************************************
 * Simple text-based program to create and manage and X.509 (aka SSL)
 * certificate authority.
 *****************************************************************************
 * Copyright (C) 2002  Jason Heiss (jheiss-javaca@ofb.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *****************************************************************************
 * $Log: JavaCA.java,v $
 * Revision 1.7  2002/10/07 03:20:17  jheiss
 * Added copyright and GPL message.
 *
 * Revision 1.6  2002/10/07 02:26:45  jheiss
 * Changed most things that were public to protected.  This is a utility,
 * not a library, and therefore really shouldn't have a bunch of public
 * stuff.
 * Rearranged methods into logical groups.
 * Broke out the X509Name generation for new certs into a seperate method.
 * Replaced generateTestCert with generatePlaceholderCert (which is all
 * the previous methods was being used for) and a generateSelfSignedCert
 * method which allows you to specify all of the important parameters.
 * Changed decryptPkcs8PrivateKey to act like it should and just decrypt
 * the key, not decode it as well.  This makes it match the behavior of
 * the OpenSSL traditional format methods as well.
 *
 * Revision 1.5  2002/10/06 03:07:11  jheiss
 * Changed package name.
 * Removed the newlines from the various HEADER constants, leaving it
 * up to the methods that use them to insert newlines where needed.
 * Changed the usage message to use the regular class path instead of
 * a jar so that I don't have to distribute a jar file and worry about
 * the classpath inside the jar.
 * Implemented listing the certificates issued.
 * Move the CA export feature to a seperate method.
 * Significant improvements to the CA import feature, which now supports
 * unencrypted PKCS#8 and OpenSSL traditional formats, as well as inserting
 * a placeholder into the KeyStore so that future certs are issued with
 * appropriate serial numbers and don't repeat previously issued certs.
 * Implemented saving the CA KeyStore back to disk after issuing a new cert
 * so that the issued cert gets recorded.
 * Removed prompting for saving new certs in OpenSSL traditional format,
 * now that I know it isn't necessary and won't be implemented.
 * Rewrote the code that sets the Issuer DN in new certs from the Subject
 * DN of the CA's cert.  Instead of converting to a String and back again
 * (which didn't work with certs originally generated by OpenSSL), it now
 * uses some BouncyCastle magic to go more directly from the JCE to the BC
 * object.
 * Added a method to generate a test certificate.
 *
 * Revision 1.4  2002/09/25 05:25:23  jheiss
 * Changed constants to use Java standard naming format.
 * Sorted methods into three major sections:  infrastructure, public and
 * protected.
 * Prompt user for keystore password instead of just hard-coding it.
 * Changed all of the menus to use the displayMenu method.
 * Changed most of the menu handling sections to switch statements instead
 * of if/else if statements.
 * Implemented exporting of the CA key.
 * Implemented importing of CA from PEM files.
 * Implemented a variety of ways of storing newly generated keys and certs.
 *
 * Revision 1.3  2002/09/20 04:05:37  jheiss
 * Cleaned up the generation of the authority key identifier, including the
 * addition of the createAuthorityKeyId method.
 *
 * Revision 1.2  2002/09/20 02:30:57  jheiss
 * Expanded java.security import
 * Fixed PKCS#8 PEM headers
 * Added -d option and debug method
 * Made getInput method protected instead of public since it isn't really
 * part of our interface
 * Added displayMenu method
 * Added makeGeneralNamePretty so that we can display the proposed subject
 * alternative name entries to the user
 * Cleaned up the generation of the subject key identifier, including the
 * addition of the createSubjectKeyId method
 *
 * Revision 1.1  2002/09/18 04:21:49  jheiss
 * JavaCA is a tool for creating and running an X.509 certificate authority.
 * It has a simple text menu interface.
 *
 *****************************************************************************
 */

package oss.javaca;

/* Imports */
import java.io.*;
// java.net
import java.net.InetAddress;
import java.net.UnknownHostException;
// java.util
import java.util.Date;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.SortedMap;
import java.util.StringTokenizer;
import java.util.TreeMap;
import java.util.Iterator;
import java.util.NoSuchElementException;
// java.text
import java.text.DateFormat;
// java.math
import java.math.BigInteger;
// java.security
import java.security.AlgorithmParameters;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Security;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.InvalidParameterException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.SignatureException;
import java.security.UnrecoverableKeyException;
// java.security.cert
import java.security.cert.Certificate;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateEncodingException;
// java.security.spec
import java.security.spec.DSAPrivateKeySpec;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.RSAPrivateCrtKeySpec;
import java.security.spec.InvalidKeySpecException;
// javax.crypto
import javax.crypto.Cipher;
import javax.crypto.EncryptedPrivateKeyInfo;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
// javax.crypto.spec
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.PBEKeySpec;
import javax.crypto.spec.PBEParameterSpec;
// javax.crypto.interfaces
import javax.crypto.interfaces.PBEKey;
// org.bouncycastle.jce
import org.bouncycastle.jce.X509V3CertificateGenerator;
// org.bouncycastle.jce.provider
import org.bouncycastle.jce.provider.BouncyCastleProvider;
// org.bouncycastle.asn1
import org.bouncycastle.asn1.DERConstructedSequence;
import org.bouncycastle.asn1.DERIA5String;
import org.bouncycastle.asn1.DERInputStream;
import org.bouncycastle.asn1.DERInteger;
import org.bouncycastle.asn1.DERObject;
import org.bouncycastle.asn1.DERObjectIdentifier;
import org.bouncycastle.asn1.DEROctetString;
import org.bouncycastle.asn1.DERTaggedObject;
// org.bouncycastle.asn1.x509
import org.bouncycastle.asn1.x509.BasicConstraints;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.GeneralNames;
import org.bouncycastle.asn1.x509.SubjectKeyIdentifier;
import org.bouncycastle.asn1.x509.SubjectPublicKeyInfo;
import org.bouncycastle.asn1.x509.AuthorityKeyIdentifier;
import org.bouncycastle.asn1.x509.X509Extensions;
import org.bouncycastle.asn1.x509.X509Name;
// org.bouncycastle.asn1.pkcs
import org.bouncycastle.asn1.pkcs.PrivateKeyInfo;
import org.bouncycastle.asn1.pkcs.RSAPrivateKeyStructure;
// org.bouncycastle.util.encoders
import org.bouncycastle.util.encoders.Base64;

public class JavaCA
{
	// These should be read from a properties file
	protected String CountryDefault = "US";
	protected String StateDefault = "Washington";
	protected String LocalityDefault = "Seattle";
	protected String OrganizationDefault = "Widgets, Inc.";
	protected String OrganizationalUnitDefault = "Engineering";
	protected long DaysValidDefault = 365;

	// The alias used for the CA key in the keystore
	// Certs are stored in the keystore based on serial number, and the
	// CA key/cert will be serial number zero.
	protected static final String CA_ALIAS = "0";

	public static final String PEM_CERTIFICATE_HEADER =
		"-----BEGIN CERTIFICATE-----";
	public static final String PEM_CERTIFICATE_FOOTER =
		"-----END CERTIFICATE-----";
	public static final String PEM_PKCS8_UNENCRYPTED_HEADER =
		"-----BEGIN PRIVATE KEY-----";
	public static final String PEM_PKCS8_UNENCRYPTED_FOOTER =
		"-----END PRIVATE KEY-----";
	public static final String PEM_PKCS8_ENCRYPTED_HEADER =
		"-----BEGIN ENCRYPTED PRIVATE KEY-----";
	public static final String PEM_PKCS8_ENCRYPTED_FOOTER =
		"-----END ENCRYPTED PRIVATE KEY-----";
	public static final String PEM_OPENSSL_RSA_HEADER =
		"-----BEGIN RSA PRIVATE KEY-----";
	public static final String PEM_OPENSSL_RSA_FOOTER =
		"-----END RSA PRIVATE KEY-----";
	public static final String PEM_OPENSSL_DSA_HEADER =
		"-----BEGIN DSA PRIVATE KEY-----";
	public static final String PEM_OPENSSL_DSA_FOOTER =
		"-----END DSA PRIVATE KEY-----";

	protected static final int KEY_ALGO_NONE = 0;
	protected static final int KEY_ALGO_RSA = 1;
	protected static final int KEY_ALGO_DSA = 2;

	protected static final int KEY_ENCODING_NONE = 0;
	protected static final int KEY_ENCODING_PKCS8 = 1;
	protected static final int KEY_ENCODING_OPENSSL = 2;

	protected static final String PLACEHOLDER_STRING = "placeholder";

	/////////////////////////////////////////////////////////////////////////
	// "Infrastructure" methods
	/////////////////////////////////////////////////////////////////////////

	public static void main(String[] args)
	{
		JavaCA ca = new JavaCA();
		ca.run(args);
	}

	public JavaCA()
	{
		// ***Load defaults from property file

		// Install the BouncyCastle provider
		// Otherwise X509V3CertificateGenerator, and anything else which
		// depends on the "BC" provider, will fail.
		Security.addProvider(new BouncyCastleProvider());
	}

	protected void usage()
	{
		System.err.println(
			//"Usage:  java -jar JavaCA-xxx.jar [-d] path/to/CA_keystore");
			"Usage:  java -cp bcprov-xxx.jar:. oss.javaca.JavaCA [-d] " +
				"path/to/CA_keystore");
		System.err.println("");
		System.err.println("The keystore doesn't have to exist yet, you will");
		System.err.println("be prompted to create it if necessary.");
		System.err.println("");
		System.err.println("The -d option turns on debug mode, which causes");
		System.err.println("stack traces for every Exception caught, and");
		System.err.println("possibly other useful debug output.");

		System.exit(0);
	}

	/////////////////////////////////////////////////////////////////////////
	// The main interactive loop
	/////////////////////////////////////////////////////////////////////////

	public void run(String[] args)
	{
		File caKeyStoreFile = null;
		KeyStore caKeyStore = null;

		//////////////////////////////////////////////////
		// Process the command line arguments
		//////////////////////////////////////////////////

		int currentArg = 0;

		boolean keystoreSpecified = false;
		while (! keystoreSpecified)
		{
			// Make sure there is still an argument left
			if (args.length <= currentArg)
			{
				usage();
			}

			if (args[currentArg].equals("-d"))
			{
				System.err.println("Debugging enabled");
				debugEnabled = true;
				currentArg++;
			}
			else
			{
				caKeyStoreFile = new File(args[currentArg]);
				keystoreSpecified = true;
			}
		}

		//////////////////////////////////////////////////
		// Initialization
		//////////////////////////////////////////////////

		// Bah, Java's platform independance bites us on the arse.
		// There's no way to read a password without echoing, short of
		// a chunk of JNI code or Swing.
		System.out.println(
			"Warning:  You password will be echoed to the screen");
		System.out.print("Password for CA keystore:  ");
		char[] password = getInput().toCharArray();

		// If the keystore doesn't exist, prompt the user about creating one
		if (! caKeyStoreFile.exists())
		{
			caKeyStore = createCAKeystore(caKeyStoreFile, password);
			if (caKeyStore == null)
			{
				System.exit(1);
			}
		}
		// Otherwise load the keystore and make sure it contains a CA key
		else
		{
			caKeyStore = validateCAKeystore(caKeyStoreFile, password);
			if (caKeyStore == null)
			{
				System.exit(1);
			}
		}

		//////////////////////////////////////////////////
		// Main loop
		//////////////////////////////////////////////////

		MAINMENU: while (true)
		{
			TreeMap mainOptions = new TreeMap();

			// Build up our menu options
			final int MAIN_ISSUE_NEW = 1;
			mainOptions.put(
				new Integer(MAIN_ISSUE_NEW),
				"Issue a new certificate");
			final int MAIN_ISSUE_ON_CSR = 2;
			mainOptions.put(
				new Integer(MAIN_ISSUE_ON_CSR),
				"#Issue a new certificate based on a CSR");
			final int MAIN_RENEW = 3;
			mainOptions.put(
				new Integer(MAIN_RENEW),
				"#Renew a previously issued certificate");
			final int MAIN_LIST = 4;
			mainOptions.put(
				new Integer(MAIN_LIST),
				"List certificates issued");
			final int MAIN_REVOKE = 5;
			mainOptions.put(
				new Integer(MAIN_REVOKE),
				"#Revoke a certificate");
			final int MAIN_EXPORT_CA = 6;
			mainOptions.put(
				new Integer(MAIN_EXPORT_CA),
				"Export CA certificate to a file");
			final int MAIN_EXIT = 7;
			mainOptions.put(
				new Integer(MAIN_EXIT),
				"Exit");

			int mainChoice = displayMenu("Main Menu", mainOptions);

			switch (mainChoice)
			{
			case MAIN_ISSUE_NEW:
				createKeyAndSignedCert(caKeyStore, caKeyStoreFile, password);
				break;
			case MAIN_ISSUE_ON_CSR:
				System.out.println("");
				System.out.println("You have chosen to issue a certificate " +
					"based on a CSR");
				System.out.println("");
				// Do we need to prompt for the CSR format (is there anything
				// but PKCS#10?) or can we figure it out automagically?
				// ***
				System.err.println("Sorry, this feature isn't implemented yet");
				break;
			case MAIN_RENEW:
				System.out.println("");
				System.out.println("You have chosen to renew a certificate");
				// ***
				System.err.println("Sorry, this feature isn't implemented yet");
				break;
			case MAIN_LIST:
				listIssuedCertificates(caKeyStore);
				break;
			case MAIN_REVOKE:
				System.out.println("");
				System.out.println("You have chosen to revoke a certificate");
				// ***
				System.err.println("Sorry, this feature isn't implemented yet");
				break;
			case MAIN_EXPORT_CA:
				exportCACertificate(caKeyStore);
				break;
			case MAIN_EXIT:
				System.out.println("Ok, exiting");
				System.exit(0);
				break;
			}
		}
	}

	/*
	 ************************************************************************
	 * Methods for handling the CA KeyStore
	 ************************************************************************
	 */

	/*
	 * This builds a new CA keystore by getting a CA key/cert from some
	 * source (the user is prompted to select a source) and then writing
	 * that into the specified keystore file using the specified password.
	 *
	 * Returns the new KeyStore if successful, null otherwise
	 */
	protected KeyStore createCAKeystore(File caKeyStoreFile, char[] password)
	{
		TreeMap createCAOptions = new TreeMap();
		final int CREATE_CA_NEW = 1;
		createCAOptions.put(
			new Integer(CREATE_CA_NEW),
			"Create a new CA");
		final int CREATE_CA_IMPORT = 2;
		createCAOptions.put(
			new Integer(CREATE_CA_IMPORT),
			"Import a CA from PEM encoded files (used by OpenSSL)");
		final int CREATE_CA_CANCEL = 3;
		createCAOptions.put(
			new Integer(CREATE_CA_CANCEL),
			"Cancel");

		System.out.println("");
		System.out.println(
			"The Certificate Authority (CA) keystore that you specified");
		System.out.println("(" + caKeyStoreFile.getName() + ") doesn't exist");

		int createCAChoice = displayMenu(
			"Options for creating a new CA keystore",
			createCAOptions);

		// Any option needs to assign to these variables
		// At the end of the method we store these into our CA keystore
		PrivateKey caKey = null;
		X509Certificate[] caCertChain = new X509Certificate[0];
		// This is the serial number that the CA should start issuing
		// certs at.  Useful when importing a CA that has already
		// issued some certs so that we don't reuse serial numbers.
		BigInteger initialSerialNumber = BigInteger.ONE;

		switch (createCAChoice)
		{
		case CREATE_CA_NEW:
			System.out.println("");
			System.out.println("You have chosen to create a new CA");

			// ***Note that both createKeyPair and createCertificate
			// currently assume RSA keys

			System.out.print("Generating key pair...");
			KeyPair keypair = createKeyPair();
			if (keypair == null)
			{
				return null;
			}
			caKey = keypair.getPrivate();
			System.out.println(" done");

			// Create a self-signed cert by specifying matching public
			// and private keys (and no CA certificate)
			System.out.println("Generating CA certificate...");
			X509Certificate newCACert = createCertificate(
				keypair.getPublic(),
				keypair.getPrivate(),
				null,
				BigInteger.valueOf(0),
				CERT_TYPE_CA);
			if (newCACert == null)
			{
				return null;
			}

			// We are the root CA in this case, so our chain is easy
			caCertChain = new X509Certificate[]{newCACert};

			break;

		case CREATE_CA_IMPORT:
			System.out.println("You have chosen to inport a CA " +
				"from OpenSSL");

			// Prompt for the CA key filename and load the key

			System.out.println("");
			System.out.println("Filename of CA key to import?");
			System.out.println(
				"This is typically like " +
				"/path/to/demoCA/private/cakey.pem");
			// ***
			//System.out.println(
				//"We support either OpenSSL traditional or " +
				//"PKCS#8 formatted files");
			System.out.println(
				"We support unencrypted PKCS#8 or OpenSSL traditional " +
				"format files.");
			System.out.println(
				"Compare the first line of your key file to this list:");
			System.out.println("");
			System.out.println(
				PEM_PKCS8_UNENCRYPTED_HEADER +
				"                PKCS#8 unencrypted");
			System.out.println(
				PEM_PKCS8_ENCRYPTED_HEADER +
				"      PKCS#8 encrypted");
			System.out.println(
				PEM_OPENSSL_RSA_HEADER +
				"            OpenSSL \"traditional\"");
			System.out.println(
				PEM_OPENSSL_DSA_HEADER +
				"            OpenSSL \"traditional\"");
			System.out.println("");
			System.out.println(
				"If the second line of your traditional format key is " +
				"'Proc-Type: 4,ENCRYPTED'");
			System.out.println(
				"then your key is encrypted.  Any of the above formats can " +
				"be converted to");
			System.out.println(
				"unencrypted PKCS#8 with this OpenSSL command:");
			System.out.println(
				"openssl pkcs8 -topk8 -in yourkey.pem -out yourkey.pk8.pem " +
				"-nocrypt");

			FileInputStream keyStream = null;
			while (keyStream == null)
			{
				String importCAKeyFilename = "";
				while (importCAKeyFilename.equals(""))
				{
					System.out.print("Filename:  ");
					importCAKeyFilename = getInput();
				}
				keyStream = openFileForRead(new File(importCAKeyFilename));
			}

			caKey = loadPEMEncodedPrivateKey(keyStream);
			try { keyStream.close(); } catch (IOException e) {}
			if (caKey == null)
			{
				return null;
			}

			// Now prompt for the CA certificate filename and load it

			System.out.println("");
			System.out.println("Filename of CA certificate to import?");
			System.out.println(
				"This is typically like /path/to/demoCA/cacert.pem");
			FileInputStream certStream = null;
			while (certStream == null)
			{
				String importCACertFilename = "";
				while (importCACertFilename.equals(""))
				{
					System.out.print("Filename:  ");
					importCACertFilename = getInput();
				}

				certStream = openFileForRead(new File(importCACertFilename));
			}

			X509Certificate importCACert = loadPEMEncodedCert(certStream);
			try { certStream.close(); } catch (IOException e) {}

			if (importCACert == null)
			{
				return null;
			}

			// ***We're just going to assume we're the root CA.  We should
			// look at the loaded cert and see if it is self-signed, and
			// if not prompt for additional certs to load.
			caCertChain = new X509Certificate[]{importCACert};

			System.out.println(
				"What is the largest serial number that has already been");
			System.out.println(
				"issued by this CA?  If this cert came from OpenSSL, you");
			System.out.println(
				"can look in /path/to/demoCA/index.txt, column three.");
			System.out.println(
				"Note that the numbers in that file are in hex.");
			boolean goodLargestSerialRead = false;
			while (! goodLargestSerialRead)
			{
				System.out.print("Largest serial [none]:  ");
				String largestSerialInput = getInput();
				if (largestSerialInput.equals(""))
				{
					// Leave initialSerialNumber set to the initial value
					goodLargestSerialRead = true;
				}
				else
				{
					try
					{
						// Add one to the largest serial number issued so far
						// to get the initial serial number
						initialSerialNumber =
							new BigInteger(largestSerialInput).
								add(BigInteger.ONE);
						// Make sure they didn't give us a negative number
						if (initialSerialNumber.compareTo(BigInteger.ZERO) <= 0)
						{
							throw new NumberFormatException();
						}
						goodLargestSerialRead = true;
					}
					catch (NumberFormatException e)
					{
						System.err.println("Invalid value entered");
					}
				}
			}

			break;
		case CREATE_CA_CANCEL:
			return null;
		}

		// Quick sanity check, we shouldn't get to this point and have
		// either one of these still at the initial values we set at the
		// beginning.
		if (caKey == null || caCertChain.length == 0)
		{
			System.err.println(
				"CA key or cert not set when we went to store them, " +
				"THIS IS A BUG!");
			return null;
		}

		// Create a new keystore and stuff our key/cert into it
		try
		{
			// Create a new, empty keystore
			KeyStore ks = KeyStore.getInstance("JKS");
			ks.load(null, null);

			// Insert the key/cert
			ks.setKeyEntry(CA_ALIAS, caKey, password, caCertChain);

			// If initialSerialNumber is greater than 1, stick a fake
			// entry in at initialSerialNumber-1 so that the next cert
			// issued will be at initialSerialNumber.
			if (initialSerialNumber.compareTo(BigInteger.ONE) > 0)
			{
				BigInteger fakeEntry =
					initialSerialNumber.subtract(BigInteger.ONE);
				ks.setCertificateEntry(fakeEntry.toString(),
					generatePlaceholderCert(fakeEntry));
			}

			// Write the keystore out to a file
			FileOutputStream kstream =
				new FileOutputStream(caKeyStoreFile);
			ks.store(kstream, password);
			kstream.close();

			return ks;
		}
		catch (KeyStoreException e)
		{
			debug(e);
			System.err.println(
				"Failed to store CA key/cert in keystore:" +
			 	e.getMessage());
			return null;
		}
		catch (FileNotFoundException e)
		{
			debug(e);
			System.err.println(
				"Failed to store CA key/cert in keystore:" +
			 	e.getMessage());
			return null;
		}
		catch (IOException e)
		{
			debug(e);
			System.err.println(
				"Failed to store CA key/cert in keystore:" +
			 	e.getMessage());
			return null;
		}
		catch (NoSuchAlgorithmException e)
		{
			debug(e);
			System.err.println(
				"Failed to store CA key/cert in keystore:" +
			 	e.getMessage());
			return null;
		}
		catch (CertificateException e)
		{
			debug(e);
			System.err.println(
				"Failed to store CA key/cert in keystore:" +
			 	e.getMessage());
			return null;
		}
	}

	/*
	 * This checks the specified keystore to make sure it contains a CA
	 * key/cert in the entry identified by the alias in the CA_ALIAS variable.
	 *
	 * Returns the loaded keystore if successful, null otherwise.
	 */
	protected KeyStore validateCAKeystore(File caKeyStoreFile, char[] password)
	{
		try
		{
			FileInputStream kstream = new FileInputStream(caKeyStoreFile);
			KeyStore ks = KeyStore.getInstance("JKS");
			ks.load(kstream, password);
			kstream.close();
			if (ks.isKeyEntry(CA_ALIAS))
			{
				// ***Do we need to do any more validation of the entry?
				// Should we check that it is self-signed?
				//   No, because we may not be a root CA.
				return ks;
			}
			else
			{
				System.err.println("The specified keystore doesn't " +
					"contain a CA key stored with the alias " + CA_ALIAS);
				return null;
			}
		}
		catch (FileNotFoundException e)
		{
			debug(e);
			System.err.println("Failed to load specified keystore:  " +
				e.getMessage());
			return null;
		}
		catch (KeyStoreException e)
		{
			debug(e);
			System.err.println("Failed to load specified keystore:  " +
				e.getMessage());
			return null;
		}
		catch (IOException e)
		{
			debug(e);
			System.err.println("Failed to load specified keystore:  " +
				e.getMessage());
			return null;
		}
		catch (NoSuchAlgorithmException e)
		{
			debug(e);
			System.err.println("Failed to load specified keystore:  " +
				e.getMessage());
			return null;
		}
		catch (CertificateException e)
		{
			debug(e);
			System.err.println("Failed to load specified keystore:  " +
				e.getMessage());
			return null;
		}
	}

	/*
	 * Of the entries in the keystore which can be converted to BigIntegers,
	 * find the maximum and return it.
	 */
	protected BigInteger getMaximumSerialNumber(KeyStore keystore)
	{
		BigInteger highest = BigInteger.valueOf(0);

		try
		{
        	for (Enumeration enum = keystore.aliases(); enum.hasMoreElements();)
        	{
				try
				{
            		BigInteger temp =
						new BigInteger((String) enum.nextElement());
					highest = highest.max(temp);
				}
				catch (NumberFormatException e)
				{
					System.err.println(
						"Warning:  Non-numeric entry in keystore");
				}
			}
		}
		catch (KeyStoreException e)
		{
			debug(e);
			System.err.println("Keystore not initialized when trying to " +
				"find maximum serial number:  " + e.getMessage());
			System.exit(1);
		}

		return highest;
	}

	/*
	 * Export the CA certificate to a file.  The user is prompted for
	 * the file format and name.
	 *
	 * Returns true if successful, false otherwise.
	 */
	protected boolean exportCACertificate(KeyStore caKeyStore)
	{
		TreeMap caExportOptions = new TreeMap();
		final int CA_EXPORT_PEM = 1;
		caExportOptions.put(
			new Integer(CA_EXPORT_PEM),
			"PEM encoded file (used by OpenSSL)");

		int caExportChoice = displayMenu(
			"You have chosen to export the CA certificate",
			caExportOptions);

		switch (caExportChoice)
		{
		case CA_EXPORT_PEM:
			FileOutputStream certStream = null;
			while (certStream == null)
			{
				System.out.print(
					"Filename for certificate [cacert.pem]:  ");
				String certFilename = getInput();
				if (certFilename.equals(""))
				{
					certFilename = "cacert.pem";
				}
				certStream = openFileForWrite(new File(certFilename));
			}

			try
			{
				return savePEMEncodedCert(
					caKeyStore.getCertificate(CA_ALIAS),
					certStream);
			}
			catch (KeyStoreException e)
			{
				debug(e);
				System.err.println(
					"Error loading the CA certificate:  " +
					e.getMessage());
				return false;
			}
		}
		return false;
	}

	/*
	 * List the issued certificates
	 *
	 * Returns true if successful, false otherwise.
	 */
	protected boolean listIssuedCertificates(KeyStore caKeyStore)
	{
		System.out.println("Certificates issued by this CA:");
		try
		{
       		for (Enumeration enum = caKeyStore.aliases();
				enum.hasMoreElements();)
			{
				String alias = (String) enum.nextElement();
				if (caKeyStore.isCertificateEntry(alias))
				{
					X509Certificate cert =
						(X509Certificate) caKeyStore.getCertificate(alias);

					// Skip the placeholder cert we put in if we imported
					// the CA
					if (cert.getSubjectX500Principal().getName().
						equals("CN=" + PLACEHOLDER_STRING))
					{
						continue;
					}

					//System.out.println(alias);
					//System.out.println("    DN: " +
					System.out.println("DN: " +
						cert.getSubjectX500Principal().getName());
					System.out.println("    Start: " +
						DateFormat.getDateInstance().format(
							cert.getNotBefore()));
					System.out.println("    End: " +
						DateFormat.getDateInstance().format(
							cert.getNotAfter()));
					// *** subjectAltName
				}
			}
			return true;
		}
		catch (KeyStoreException e)
		{
			debug(e);
			System.out.println("Error reading certs:  " + e.getMessage());
			return false;
		}
	}

	/*
	 ************************************************************************
	 * Methods for generating keys and certificates
	 ************************************************************************
	 */

	/*
	 * Create a public/private key pair and sign the public key (thus making
	 * a certificate) using the CA key from the specified keystore.
	 * Then prompt the user for a file format and save the new key and cert.
	 *
	 * This should probably just take the File and password and call
	 * validateCAKeystore to load the KeyStore.
	 *
	 * Returns true if successful, false otherwise.
	 */
	protected boolean createKeyAndSignedCert(
		KeyStore caKeyStore, File caKeyStoreFile, char[] password)
	{
		KeyPair keypair = null;
		X509Certificate cert = null;

		TreeMap genCertOptions = new TreeMap();
		final int GEN_CERT_HOST = 1;
		genCertOptions.put(
			new Integer(GEN_CERT_HOST),
			"Generate a host certificate");
		final int GEN_CERT_USER = 2;
		genCertOptions.put(
			new Integer(GEN_CERT_USER),
			"Generate a user certificate");
		final int GEN_CERT_GENERIC = 3;
		genCertOptions.put(
			new Integer(GEN_CERT_GENERIC),
			"Generate a generic certificate");
		final int GEN_CERT_CANCEL = 4;
		genCertOptions.put(
			new Integer(GEN_CERT_CANCEL),
			"Cancel");

		int genCertChoice = displayMenu(
			"You have chosen to generate a new certificate",
			genCertOptions);

		if (genCertChoice == GEN_CERT_CANCEL)
		{
			// I guess this can be considered successful...
			return true;
		}

		//////////////////////////////////////////////////
		// Generate the keys and certificate
		//////////////////////////////////////////////////

		System.out.print("Generating key pair...");
		keypair = createKeyPair();
		if (keypair == null)
		{
			return false;
		}
		System.out.println(" done");

		System.out.println("Generating certificate...");

		int certType = CERT_TYPE_GENERIC;
		if (genCertChoice == GEN_CERT_HOST)
		{
			certType = CERT_TYPE_HOST;
		}
		else if (genCertChoice == GEN_CERT_USER)
		{
			certType = CERT_TYPE_USER;
		}

		// Load the CA key and cert from the keystore for
		// passing along to createCertificate()
		PrivateKey caKey = null;
		X509Certificate caCert = null;
		try
		{
			caKey = (PrivateKey) caKeyStore.getKey(CA_ALIAS, password);
			caCert = (X509Certificate)
				caKeyStore.getCertificate(CA_ALIAS);
		}
		catch(KeyStoreException e)
		{
			debug(e);
			System.err.println("Error loading CA key from keystore:  " +
				e.getMessage());
			return false;
		}
		catch(NoSuchAlgorithmException e)
		{
			debug(e);
			System.err.println("Error loading CA key from keystore:  " +
				e.getMessage());
			return false;
		}
		catch(UnrecoverableKeyException e)
		{
			debug(e);
			System.err.println("Error loading CA key from keystore:  " +
				e.getMessage());
			return false;
		}

		// Create the certificate
		cert = createCertificate(
			keypair.getPublic(),
			caKey,
			caCert,
			getMaximumSerialNumber(caKeyStore).add(BigInteger.ONE),
			certType);

		// Store the newly-minted certificate in our CA KeyStore
		// so that we can keep track of certs we've issued.  This
		// might be somewhat of a misuse of a KeyStore, but
		// CertStore doesn't define/provide a file-based storage
		// mechanism.
		try
		{
			caKeyStore.setCertificateEntry(
				cert.getSerialNumber().toString(), cert);
			
			// Save the CA KeyStore back to disk
			FileOutputStream caStream = openFileForWrite(caKeyStoreFile);
			caKeyStore.store(caStream, password);
			caStream.close();
		}
		catch (KeyStoreException e)
		{
			debug(e);
			System.err.println("Error saving new cert to CA KeyStore:  " +
				e.getMessage());
			System.exit(1);
		}
		catch (IOException e)
		{
			debug(e);
			System.err.println("Error saving new cert to CA KeyStore:  " +
				e.getMessage());
			System.exit(1);
		}
		catch (NoSuchAlgorithmException e)
		{
			debug(e);
			System.err.println("Error saving new cert to CA KeyStore:  " +
				e.getMessage());
			System.exit(1);
		}
		catch (CertificateException e)
		{
			debug(e);
			System.err.println("Error saving new cert to CA KeyStore:  " +
				e.getMessage());
			System.exit(1);
		}

		//////////////////////////////////////////////////
		// Now that we've generated a key and cert,
		// we need to save them to a file(s) for the user.
		//////////////////////////////////////////////////

		// Prompt the user for the key and certificate file format
		TreeMap storeCertOptions = new TreeMap();
		final int CERT_STORE_PEM = 1;
		storeCertOptions.put(
			new Integer(CERT_STORE_PEM),
			"PEM encoded files (used by OpenSSL)");
		final int CERT_STORE_PKCS12 = 2;
		storeCertOptions.put(
			new Integer(CERT_STORE_PKCS12),
			"PKCS#12 file (used by web browsers)");
		final int CERT_STORE_KEYSTORE = 3;
		storeCertOptions.put(
			new Integer(CERT_STORE_KEYSTORE),
			"Java keystore");

		int storeCertChoice = displayMenu(
			"How would you like to store the key and certificate:",
			storeCertOptions);

		switch (storeCertChoice)
		{
		case CERT_STORE_PEM:

			// Prompt the user for the private key file format
			TreeMap keyFormatOptions = new TreeMap();
			final int KEY_FORMAT_PKCS8_UNENC = 1;
			keyFormatOptions.put(
				new Integer(KEY_FORMAT_PKCS8_UNENC),
				"PKCS#8 unencrypted");
			final int KEY_FORMAT_PKCS8_ENC = 2;
			keyFormatOptions.put(
				new Integer(KEY_FORMAT_PKCS8_ENC),
				"#PKCS#8 encrypted");

			int keyFormatChoice = displayMenu(
				"How would you like to store the private key:",
				keyFormatOptions);

			// Loop until the user gives us a file we can open
			// for the private key
			FileOutputStream keyStream = null;
			while (keyStream == null)
			{
				System.out.print(
					"Filename for private key [newkey.pem]:  ");
				String keyFilename = getInput();
				if (keyFilename.equals(""))
				{
					keyFilename = "newkey.pem";
				}
				keyStream = openFileForWrite(new File(keyFilename));
			}

			switch (keyFormatChoice)
			{
			case KEY_FORMAT_PKCS8_UNENC:
				savePKCS8EncodedKey(keypair.getPrivate(), keyStream);
				break;
			case KEY_FORMAT_PKCS8_ENC:
				System.err.println("Sorry, this feature " +
					"isn't implemented yet");
				break;
			}
			
			// Loop until the user gives us a file we can open
			// for the certificate
			FileOutputStream certStream = null;
			while (certStream == null)
			{
				System.out.print(
					"Filename for certificate [newcert.pem]:  ");
				String certFilename = getInput();
				if (certFilename.equals(""))
				{
					certFilename = "newcert.pem";
				}
				certStream = openFileForWrite(new File(certFilename));
			}
			savePEMEncodedCert(cert, certStream);

			return true;

		// The BouncyCastle provider supports loading and storing PKCS#12
		// files via the KeyStore interface (the Sun provider can only
		// load PKCS#12 files).
		//
		// So we can support both PKCS#12 and the proprietary "JKS" Java
		// keystore with mostly the same hunk of code, just modifying our
		// prompts a little bit, and of course changing the type of
		// keystore we ask for with KeyStore.getInstance()
		case CERT_STORE_PKCS12:
		case CERT_STORE_KEYSTORE:
			// Loop until the user gives us a file we can open
			// for the keystore
			FileOutputStream keystoreStream = null;
			while (keystoreStream == null)
			{
				String defaultFilename = "newfile";
				if (storeCertChoice == CERT_STORE_PKCS12)
				{
					defaultFilename = "new.pkcs12";
					System.out.print(
						"Filename for PKCS#12 file [" + defaultFilename +
						"]:  ");
				}
				else if (storeCertChoice == CERT_STORE_KEYSTORE)
				{
					defaultFilename = "new.keystore";
					System.out.print(
						"Filename for Java keystore [" + defaultFilename +
						"]:  ");
				}
				String keystoreFilename = getInput();
				if (keystoreFilename.equals(""))
				{
					keystoreFilename = defaultFilename;
				}
				keystoreStream = openFileForWrite(new File(keystoreFilename));
			}

			// Loop until the user gives us a password for the new keystore
			String newKeystorePassword = "";
			while (newKeystorePassword.equals(""))
			{
				if (storeCertChoice == CERT_STORE_PKCS12)
				{
					System.out.print("Password for PKCS#12 file:  ");
				}
				else if (storeCertChoice == CERT_STORE_KEYSTORE)
				{
					System.out.print("Password for Java keystore:  ");
				}
				newKeystorePassword = getInput();
				if (newKeystorePassword.equals(""))
				{
					System.err.println("Password can't be blank");
				}
			}

			// Loop until the user gives us an alias under which to store
			// the key and certificate in the keystore
			String newAlias = "";
			while (newAlias.equals(""))
			{
				if (storeCertChoice == CERT_STORE_PKCS12)
				{
					System.out.print("Friendly name to assign to key:  ");
				}
				else if (storeCertChoice == CERT_STORE_KEYSTORE)
				{
					System.out.print(
						"Alias under which to store key and cert:  ");
				}
				newAlias = getInput();
				if (newAlias.equals(""))
				{
					System.err.println("Name must be specified");
				}
			}

			try
			{
				KeyStore newKS = null;
				// Create a new KeyStore
				if (storeCertChoice == CERT_STORE_PKCS12)
				{
					// Request a PKCS#12 keystore using the Bouncy Castle
					// provider
					newKS = KeyStore.getInstance("PKCS12", "BC");
				}
				else if (storeCertChoice == CERT_STORE_KEYSTORE)
				{
					newKS = KeyStore.getInstance("JKS");
				}
				// Initialize it
				newKS.load(null, null);
				// Retrieve the CA's cert chain
				Certificate[] caChain =
					caKeyStore.getCertificateChain(CA_ALIAS);
				// Add our certificate to the chain
				Certificate[] newChain = new Certificate[caChain.length + 1];
				System.arraycopy(caChain, 0, newChain, 0, caChain.length);
				newChain[caChain.length] = cert;
				// Store the key and cert
				newKS.setKeyEntry(
					newAlias,
					keypair.getPrivate(),
					newKeystorePassword.toCharArray(),
					newChain);
				// Save the keystore
				newKS.store(keystoreStream, newKeystorePassword.toCharArray());
				keystoreStream.close();

				return true;
			}
			catch(NoSuchProviderException e)
			{
				debug(e);
				System.err.println(
					"Error saving new key and certificate to keystore:  " +
					e.getMessage());
				return false;
			}
			catch(KeyStoreException e)
			{
				debug(e);
				System.err.println(
					"Error saving new key and certificate to keystore:  " +
					e.getMessage());
				return false;
			}
			catch(IOException e)
			{
				debug(e);
				System.err.println(
					"Error saving new key and certificate to keystore:  " +
					e.getMessage());
				return false;
			}
			catch(NoSuchAlgorithmException e)
			{
				debug(e);
				System.err.println(
					"Error saving new key and certificate to keystore:  " +
					e.getMessage());
				return false;
			}
			catch(CertificateException e)
			{
				debug(e);
				System.err.println(
					"Error saving new key and certificate to keystore:  " +
					e.getMessage());
				return false;
			}
		}

		return false;
	}

	// Creates and returns a public/private pair of keys
	// Returns null if there is an error
	protected KeyPair createKeyPair()
	{
		// ***We should prompt for these
		String keyType = "RSA";
		int keySize = 1024;

		KeyPair keypair = null;
		try
		{
			KeyPairGenerator keyGen = KeyPairGenerator.getInstance(keyType);
			keyGen.initialize(keySize);
			keypair = keyGen.generateKeyPair();
		}
		catch (NoSuchAlgorithmException e)
		{
			debug(e);
			System.err.println(
				"Failed to generate a pair of " + keyType + " keys:  " +
				e.getMessage());
			return null;
		}
		catch (InvalidParameterException e)
		{
			debug(e);
			System.err.println(
				"Failed to generate a pair of " + keyType + " keys:  " +
				e.getMessage());
			return null;
		}

		return keypair;
	}

	protected static final int CERT_TYPE_GENERIC = 0;
	protected static final int CERT_TYPE_CA = 1;
	protected static final int CERT_TYPE_HOST = 2;
	protected static final int CERT_TYPE_USER = 3;
	protected X509Certificate createCertificate(
		PublicKey keyToCertify,  // The key to make into a certificate
		PrivateKey signingKey,  // The key to sign with (typically the CA's key)
		X509Certificate signingCert,  // The cert matching signingKey
		BigInteger serialNumber,  // The serial number for the new certificate
		int certType)
	{
		String country = "";
		String state = "";
		String locality = "";
		String organization = "";
		String organizationalUnit = "";
		String commonName = "";
		// Vector of org.bouncycastle.asn1.x509.GeneralName's
		Vector subjectAltName = new Vector();
		long daysValid = 0;
		Date startDate = null;
		Date endDate = null;

		boolean userIsHappyWithInput = false;
		while (! userIsHappyWithInput)
		{
			System.out.print(
				"Country (2 letter code) [" + CountryDefault + "]:  ");
			country = getInput();
			if (country.equals(""))
			{
				country = CountryDefault;
			}

			System.out.print(
				"State or Province (full name) [" + StateDefault + "]:  ");
			state = getInput();
			if (state.equals(""))
			{
				state = StateDefault;
			}

			System.out.print(
				"Locality (eg, city) [" + LocalityDefault + "]:  ");
			locality = getInput();
			if (locality.equals(""))
			{
				locality = LocalityDefault;
			}

			System.out.print(
				"Organization (eg, company) [" + OrganizationDefault + "]:  ");
			organization = getInput();
			if (organization.equals(""))
			{
				organization = OrganizationDefault;
			}

			System.out.print(
				"Organizational unit [" + OrganizationalUnitDefault + "]:  ");
			organizationalUnit = getInput();
			if (organizationalUnit.equals(""))
			{
				organizationalUnit = OrganizationalUnitDefault;
			}

			// No default here since it will change with each cert
			while (commonName.equals(""))
			{
				System.out.print("Common name");
				if (certType == CERT_TYPE_CA)
				{
					System.out.print(" (eg, Widgets Certficate Authority)");
				}
				else if (certType == CERT_TYPE_HOST)
				{
					System.out.print(
						" (server's canonical hostname, i.e. foo.widgets.com)");
				}
				else if (certType == CERT_TYPE_USER)
				{
					System.out.print(" (User's full name, i.e. Joe User)");
				}
				System.out.print(":  ");
				commonName = getInput();
			}

			// Based on the type of cert we're generating, configure the
			// subjectAltName X.509 version 3 extension field.
			//
			// Valid values and proper usage of this field are documented
			// in RFC 2459.  http://www.ietf.org/rfc/rfc2459.txt
			if (certType == CERT_TYPE_HOST)
			{
				// Add the main hostname to subjectAltName
				subjectAltName.addElement(
					new GeneralName(new DERIA5String(commonName), 2));

				// Prompt for additional hostnames to add to subjectAltName
				boolean noMoreHostnames = false;
				while (! noMoreHostnames)
				{
					System.out.print("Additional hostname to add to cert " +
						"(enter for none):  ");
					String addnHostnameInput = getInput();
					if (addnHostnameInput.equals(""))
					{
						noMoreHostnames = true;
					}
					else
					{
						subjectAltName.addElement(
							new GeneralName(
								new DERIA5String(addnHostnameInput), 2));
					}
				}
			}
			else if (certType == CERT_TYPE_USER)
			{
				// Store the user's email address in the subjectAltName field
				System.out.print("Email address:  ");
				String emailAddress = getInput();
				if (! emailAddress.equals(""))
				{
					subjectAltName.addElement(
						new GeneralName(new DERIA5String(emailAddress), 1));
				}
			}

			System.out.print(
				"How long (in days) should this certificate be " +
				"valid [" + DaysValidDefault + "]:  ");
			String daysValidString = getInput();
			if (daysValidString.equals(""))
			{
				daysValid = DaysValidDefault;
			}
			else
			{
				try
				{
					daysValid = Long.parseLong(daysValidString);
					if (daysValid <= 0)
					{
						throw new NumberFormatException();
					}
				}
				catch (NumberFormatException e)
				{
					System.err.println(
						"Invalid value entered, reverting to default");
					daysValid = DaysValidDefault;
				}
			}
			startDate = new Date();
			//                  ms/s   s/m  m/h  h/d
			long millisPerDay = 1000 * 60 * 60 * 24;
			endDate = new Date(startDate.getTime() +
				(daysValid * millisPerDay));

			// Confirm with the user
			System.out.println("\nPlease confirm your input:");
			System.out.println("");
			System.out.println("Country:  " + country);
			System.out.println("State or province:  " + state);
			System.out.println("Locality:  " + locality);
			System.out.println("Organization:  " + organization);
			System.out.println("Organizational Unit:  " +
				organizationalUnit);
			System.out.println("Common name:  " + commonName);
			System.out.println("Subject Alternative Name:  ");
        	for (Enumeration enum = subjectAltName.elements();
				enum.hasMoreElements();)
			{
				System.out.println("    " +
					makeGeneralNamePretty((GeneralName) enum.nextElement()));
			}
			System.out.println("Valid:  " +
				daysValid + " days" +
				", Start: " + DateFormat.getDateInstance().format(startDate) +
				", End: " + DateFormat.getDateInstance().format(endDate));

			System.out.println("");
			System.out.print("Continue?  [Y/n]:  ");

			// ***We should give the user the option to cancel
			String input = getInput();
			if (input.equals("") || input.compareToIgnoreCase("y") == 0)
			{
				userIsHappyWithInput = true;
			}
			else
			{
				// User is not satisfied, reset defaults to the
				// values he's entered so far and loop back
				CountryDefault = country;
				StateDefault = state;
				LocalityDefault = locality;
				OrganizationDefault = organization;
				OrganizationalUnitDefault = organizationalUnit;

				subjectAltName.removeAllElements();
			}
		}
		
		X509Name certDN = generateX509Name(
			country,
			state,
			locality,
			organization,
			organizationalUnit,
			commonName);

		// Create our certificate generator
		X509V3CertificateGenerator certGen = new X509V3CertificateGenerator();

		// Fill in all of the fields
		if (certType == CERT_TYPE_CA)
		{
			certGen.setIssuerDN(certDN);
		}
		else
		{
			// Get the CA's DN for our issuer DN
			// It takes a little trickery to go from a JCE object
			// to a BouncyCastle object.
			try
			{
				// Make a stream out of the DER encoded DN
				DERInputStream dnStream =
                	new DERInputStream(
						new ByteArrayInputStream(
							signingCert.getSubjectX500Principal().
								getEncoded()));
				// Read the DN from the stream
				DERConstructedSequence dnSequence =
					(DERConstructedSequence) dnStream.readObject();
				// Add it to the cert generator
				certGen.setIssuerDN(new X509Name(dnSequence));
			}
			catch (IOException e)
			{
				debug(e);
				System.err.println("Error when retrieving DN from CA cert:  " +
					e.getMessage());
				return null;
			}
		}
		certGen.setSubjectDN(certDN);
		if (subjectAltName.size() != 0)
		{
			DERConstructedSequence gnSequence = new DERConstructedSequence();
        	for (Enumeration enum = subjectAltName.elements();
				enum.hasMoreElements();)
			{
				gnSequence.addObject((GeneralName) enum.nextElement());
			}

			certGen.addExtension(X509Extensions.SubjectAlternativeName,
				false, new GeneralNames(gnSequence));
		}
		if (certType == CERT_TYPE_CA)
		{
			// As required by section 4.2.1.10 of RFC 2459
			// Just as a note, OpenSSL does not set this as critical
			certGen.addExtension(X509Extensions.BasicConstraints, true,
				new BasicConstraints(true));
		}
		certGen.addExtension(X509Extensions.SubjectKeyIdentifier, false,
			createSubjectKeyId(keyToCertify));
		if (certType != CERT_TYPE_CA)
		{
			certGen.addExtension(X509Extensions.AuthorityKeyIdentifier, false,
				createAuthorityKeyId(signingCert.getPublicKey()));
		}
		// ***We should also be setting the Key Usage extension
		certGen.setNotBefore(startDate);
		certGen.setNotAfter(endDate);
		certGen.setSerialNumber(serialNumber);
		// ***This should be prompted for or user-settable somehow.  It
		// needs to be based off of what createKeyPair does.
		// See http://www.bouncycastle.org/specifications.html for
		// supported signature algorithms.
		certGen.setSignatureAlgorithm("MD5withRSA");
		certGen.setPublicKey(keyToCertify);

		// Generate a certificate using the key provided
		X509Certificate cert = null;
		try
		{
			cert = certGen.generateX509Certificate(signingKey);
		}
		catch (SecurityException e)
		{
			debug(e);
			System.err.println("Error when generating certificate:  " +
				e.getMessage());
			return null;
		}
		catch (SignatureException e)
		{
			debug(e);
			System.err.println("Error when generating certificate:  " +
				e.getMessage());
			return null;
		}
		catch (InvalidKeyException e)
		{
			debug(e);
			System.err.println("Error when generating certificate:  " +
				e.getMessage());
			return null;
		}

		return cert;
	}

	/*
	 * Generate an X509Certificate with a DN of "CN=placeholder"
	 *
	 * Returns the cert if successful, null otherwise.
	 */
	protected X509Certificate generatePlaceholderCert(BigInteger serialNumber)
	{
		return generateSelfSignedCert(
			createKeyPair(),
			generateX509Name(null, null, null, null, null, PLACEHOLDER_STRING),
			new Date(),
			new Date(),
			serialNumber);
	}

	/*
	 * Generate a self-signed X509Certificate with the specified settings.
	 *
	 * Returns the cert if successful, null otherwise.
	 */
	protected X509Certificate generateSelfSignedCert(
		KeyPair keys,
		X509Name certDN,
		Date notBefore,
		Date notAfter,
		BigInteger serialNumber)
	{
		X509V3CertificateGenerator certGen = new X509V3CertificateGenerator();
		certGen.setIssuerDN(certDN);
		certGen.setSubjectDN(certDN);
		certGen.addExtension(X509Extensions.SubjectKeyIdentifier, false,
			createSubjectKeyId(keys.getPublic()));
		certGen.addExtension(X509Extensions.AuthorityKeyIdentifier, false,
			createAuthorityKeyId(keys.getPublic()));
		certGen.setNotBefore(notBefore);
		certGen.setNotAfter(notAfter);
		certGen.setSerialNumber(serialNumber);
		certGen.setSignatureAlgorithm("MD5withRSA");
		certGen.setPublicKey(keys.getPublic());
		try
		{
			return certGen.generateX509Certificate(keys.getPrivate());
		}
		catch (SecurityException e)
		{
			debug(e);
			System.err.println("Error when generating certificate:  " +
				e.getMessage());
			return null;
		}
		catch (SignatureException e)
		{
			debug(e);
			System.err.println("Error when generating certificate:  " +
				e.getMessage());
			return null;
		}
		catch (InvalidKeyException e)
		{
			debug(e);
			System.err.println("Error when generating certificate:  " +
				e.getMessage());
			return null;
		}
	}

	/*
	 * The SubjectKeyIdentifier is basically a hash of the public
	 * key in the certificate.
	 * See section 4.2.1.2 of RFC 2459
	 */
	protected SubjectKeyIdentifier createSubjectKeyId(PublicKey pubKey)
	{
		SubjectKeyIdentifier subjectKeyId = null;

		try
		{
			ByteArrayInputStream bIn =
				new ByteArrayInputStream(pubKey.getEncoded());
			SubjectPublicKeyInfo info = new SubjectPublicKeyInfo(
				(DERConstructedSequence) new DERInputStream(bIn).readObject());
			subjectKeyId = new SubjectKeyIdentifier(info);
		}
		catch (IOException e)
		{
			debug(e);
			System.err.println("Error generating SubjectKeyIdentifier:  " +
				e.getMessage());
			System.exit(1);
		}

		return subjectKeyId;
	}

	/*
	 * The AuthorityKeyIdentifier is the SubjectKeyIdentifier from the
	 * CA's certificate.
	 * See section 4.2.1.1 of RFC 2459
	 */
	protected AuthorityKeyIdentifier createAuthorityKeyId(PublicKey pubKey)
	{
		AuthorityKeyIdentifier authKeyId = null;

		try
		{
			ByteArrayInputStream bIn =
				new ByteArrayInputStream(pubKey.getEncoded());
			SubjectPublicKeyInfo info = new SubjectPublicKeyInfo(
				(DERConstructedSequence)new DERInputStream(bIn).readObject());
			authKeyId = new AuthorityKeyIdentifier(info);
		}
		catch (IOException e)
		{
			debug(e);
			System.err.println("Error generating SubjectKeyIdentifier:  " +
				e.getMessage());
			System.exit(1);
		}

		return authKeyId;
	}

	/*
	 ************************************************************************
	 * Methods for handling Distinguished Names (DNs)
	 ************************************************************************
	 */

	/*
	 * Creates an X509Name from the specified DN elements.  Any element
	 * may be omitted by supplying null for that value.  Just make sure
	 * you supply at least one.
	 *
	 * Returns the X509Name if successful, null otherwise.
	 */
	protected X509Name generateX509Name(
		String country,
		String state,
		String locality,
		String organization,
		String organizationalUnit,
		String commonName)
	{
		Vector oids = new Vector();
		Vector attributes = new Vector();

		if (country != null)
		{
			oids.addElement(X509Name.C);
			attributes.addElement(country);
		}
		if (state != null)
		{
			oids.addElement(X509Name.ST);
			attributes.addElement(state);
		}
		if (locality != null)
		{
			oids.addElement(X509Name.L);
			attributes.addElement(locality);
		}
		if (organization != null)
		{
			oids.addElement(X509Name.O);
			attributes.addElement(organization);
		}
		if (organizationalUnit != null)
		{
			oids.addElement(X509Name.OU);
			attributes.addElement(organizationalUnit);
		}
		if (commonName != null)
		{
			oids.addElement(X509Name.CN);
			attributes.addElement(commonName);
		}

		// Make sure they gave us at least one element
		if (oids.size() > 0)
		{
			return new X509Name(oids, attributes);
		}
		else
		{
			return null;
		}
	}

	/*
	 ************************************************************************
	 * Methods for file and user I/O
	 ************************************************************************
	 */

	/*
	 * Any serious exceptions caught in this program should be passed to
	 * this method (preferably before you do anything else in your catch
	 * block, particularly anything else that prints to the screen).
	 *
	 * This method will attempt to do something useful if the user has
	 * requested debugging messages.
	 */
	protected boolean debugEnabled = false;
	protected void debug(Exception e)
	{
		if (debugEnabled)
		{
			e.printStackTrace();
		}
	}

	/*
	 * This method reads a line of input from System.in and returns it
	 */
	protected BufferedReader br = null;
	protected String getInput()
	{
		// If we haven't already initialized br
		if (br == null)
		{
			br = new BufferedReader(new InputStreamReader(System.in));
		}

		String input = "";
		try
		{
			input = br.readLine();
		}
		catch (IOException e)
		{
			debug(e);
			System.err.println("Error reading your input:  " + e.getMessage());
			System.exit(1);
		}

		return input;
	}

	/*
	 * Display a list of options to the user, prompt them for input, and
	 * returns the number of the selected option.
	 *
	 * For example:
	 * int OPTION_ONE = 1;
	 * options.put(new Integer(OPTION_ONE), "First option");
	 * int OPTION_TWO = 2;
	 * options.put(new Integer(OPTION_TWO), "Second option");
	 * displayMenu("My Menu", options, "Select: ");
	 *
	 * displays:
	 *
	 * My Menu
	 *
	 * 1:  First option
	 * 2:  Second option
	 *
	 * Select:
	 *
	 * If title is null it will be omitted.
	 * If prompt is null a default will be used.
	 */
	protected int displayMenu(String title, SortedMap options, String prompt)
	{
		if (prompt == null)
		{
			prompt = "Choice: ";
		}

		// Display title (if provided)
		if (title != null)
		{
			System.out.println("");
			System.out.println(title);
			System.out.println("");
		}

		// Display options
       	for (Iterator iter = options.keySet().iterator(); iter.hasNext();)
		{
			Integer key = (Integer) iter.next();
			System.out.println(key + ":  " + (String) options.get(key));
		}

		System.out.println("");

		boolean validInput = false;
		Integer optionNumber = null;
		while (! validInput)
		{
			// Display the prompt and wait for input
			System.out.print(prompt);
			String input = getInput();

			// Verify the user gave us a valid option number
			try
			{
				optionNumber = new Integer(input);
				if (options.containsKey(optionNumber))
				{
					validInput = true;
				}
				else
				{
					// The cheap way to only specify the error message once,
					// we should throw something related to invalid input.
					throw new NumberFormatException();
				}
			}
			catch (NumberFormatException e)
			{
				System.out.println("That is not a valid option");
			}
		}
		return optionNumber.intValue();
	}

	/*
	 * Same as the three option displayMenu but with a default prompt.
	 */
	protected int displayMenu(String title, SortedMap options)
	{
		return displayMenu(title, options, null);
	}

	/*
	 * Attempts to open the specified file as a FileOutputStream.
	 * Returns the stream if successful.
	 * Prints a warning to System.err and returns null if not successful.
	 */
	protected FileOutputStream openFileForWrite(File fileToOpen)
	{
		FileOutputStream fstream = null;
		try
		{
			fstream = new FileOutputStream(fileToOpen);
		}
		catch (FileNotFoundException e)
		{
			System.err.println("Unable to open file for writing");
		}
		return fstream;
	}

	/*
	 * Attempts to open the specified file as a FileInputStream.
	 * Returns the stream if successful.
	 * Prints a warning to System.err and returns null if not successful.
	 */
	protected FileInputStream openFileForRead(File fileToOpen)
	{
		FileInputStream fstream = null;
		try
		{
			fstream = new FileInputStream(fileToOpen);
		}
		catch (FileNotFoundException e)
		{
			System.err.println("Unable to open file for reading");
		}
		return fstream;
	}

	/*
	 ************************************************************************
	 * Methods for loading and saving key and certificate files
	 ************************************************************************
	 */

	/*
	 * Attempts to load a private key from the specified input stream.
	 *
	 * Eventually:
	 * RSA and DSA keys in either PKCS#8 or OpenSSL "traditional" format,
	 * encrypted or unencrypted, are supported (and detected automatically).
	 *
	 * Right now:
	 * RSA keys in PKCS#8 and RSA or DSA keys in OpenSSL format,
	 * unencrypted only, are supported.
	 *
	 * Returns the key if successful, null otherwise.
	 */
	protected PrivateKey loadPEMEncodedPrivateKey(InputStream keyStream)
	{
		PrivateKey key = null;

		int keyAlgo = KEY_ALGO_NONE;
		int keyEncoding = KEY_ENCODING_NONE;
		boolean keyEncrypted = false;

		String openSSLEncryptionAlgorithm = null;
		byte[] openSSLSalt = null;

		String keyText = "";

		try
		{
			// Read the first line to find out what type of key it is
			BufferedReader keyBR =
				new BufferedReader(new InputStreamReader(keyStream));

			String keyFirstLine = keyBR.readLine();
			if (keyFirstLine == null)
			{
				System.err.println("Specified file is empty");
			}
			else if (keyFirstLine.equals(PEM_PKCS8_UNENCRYPTED_HEADER))
			{
				System.out.println("Key is PKCS#8, unencrypted");
				keyEncoding = KEY_ENCODING_PKCS8;
			}
			else if (keyFirstLine.equals(PEM_PKCS8_ENCRYPTED_HEADER))
			{
				System.out.println("Key is PKCS#8, encrypted");
				keyEncoding = KEY_ENCODING_PKCS8;
				keyEncrypted = true;
			}
			else if (keyFirstLine.equals(PEM_OPENSSL_RSA_HEADER))
			{
				System.out.println("Key is OpenSSL traditional, RSA");
				keyEncoding = KEY_ENCODING_OPENSSL;
				keyAlgo = KEY_ALGO_RSA;
			}
			else if (keyFirstLine.equals(PEM_OPENSSL_DSA_HEADER))
			{
				System.out.println("Key is OpenSSL traditional, DSA");
				keyEncoding = KEY_ENCODING_OPENSSL;
				keyAlgo = KEY_ALGO_DSA;
			}

			if (keyEncoding == KEY_ENCODING_NONE)
			{
				return null;
			}

			// Load the rest of the key
			boolean firstLine = true;
			boolean openSSLHeaders = false;
			boolean endOfKeyReached = false;
			while (! endOfKeyReached)
			{
				String line = keyBR.readLine();
				if (line == null)
				{
					System.err.println(
						"End of file reached without seeing " +
						"standard footer");
					endOfKeyReached = true;
				}
				else if (
					keyEncoding == KEY_ENCODING_PKCS8 &&
					! keyEncrypted &&
					line.equals(PEM_PKCS8_UNENCRYPTED_FOOTER))
				{
					endOfKeyReached = true;
				}
				else if (
					keyEncoding == KEY_ENCODING_PKCS8 &&
					keyEncrypted &&
					line.equals(PEM_PKCS8_ENCRYPTED_FOOTER))
				{
					endOfKeyReached = true;
				}
				else if (
					keyEncoding == KEY_ENCODING_OPENSSL &&
					keyAlgo == KEY_ALGO_RSA &&
					line.equals(PEM_OPENSSL_RSA_FOOTER))
				{
					endOfKeyReached = true;
				}
				else if (
					keyEncoding == KEY_ENCODING_OPENSSL &&
					keyAlgo == KEY_ALGO_DSA &&
					line.equals(PEM_OPENSSL_DSA_FOOTER))
				{
					endOfKeyReached = true;
				}
				// Details can be found in RFC 1421 (PEM)
				// Specifically section 4.6.1
				//
				// Also helpful if you're trying to figure this section
				// out would be to look at some sample keys as generated
				// by OpenSSL.
				else if (
					keyEncoding == KEY_ENCODING_OPENSSL &&
					firstLine &&
					line.startsWith("Proc-Type:"))
				{
					openSSLHeaders = true;

					System.out.println("OpenSSL format key is encrypted");
					if (line.equals("Proc-Type: 4,ENCRYPTED"))
					{
						keyEncrypted = true;
					}
				}
				else if (
					openSSLHeaders &&
					keyEncrypted &&
					line.startsWith("DEK-Info:"))
				{
					// Sample line:
					// DEK-Info: DES-EDE3-CBC,51BFB8DFA28FC943

					try
					{
						// Hack out the "DEK-Info: "
						StringTokenizer stSpace = new StringTokenizer(line);
						stSpace.nextToken();

						// Now make a StringTokenizer based on the rest of
						// the line, using a comma as the delimeter.
						StringTokenizer stComma =
							new StringTokenizer(
							stSpace.nextToken(), ",");

						// Get the encryption algorithm
						openSSLEncryptionAlgorithm = stComma.nextToken();
						System.out.println("Key is encrypted using " +
							openSSLEncryptionAlgorithm);

						// And the salt
						String saltString = stComma.nextToken();
						// The salt is stored as a string of hex
						openSSLSalt =
							new BigInteger(saltString, 16).toByteArray();
					}
					catch (NoSuchElementException e)
					{
						debug(e);
						System.err.println(
							"DEK-Info line in PEM encoded key is " +
							"malformed");
						return null;
					}
					catch (NumberFormatException e)
					{
						debug(e);
						System.err.println(
							"DEK-Info line in PEM encoded key is " +
							"malformed");
						return null;
					}
				}
				else if (openSSLHeaders && line.equals(""))
				{
					// This signifies the end of the headers
					openSSLHeaders = false;
				}
				else if (! openSSLHeaders)
				{
					keyText = keyText + line;
				}
				else
				{
					System.err.println("Unrecognized line in key:  " + line);
				}

				firstLine = false;
			}
		}
		catch (IOException e)
		{
			debug(e);
			System.err.println("Error reading key:  " + e.getMessage());
			return null;
		}

		// Return if we didn't find any key text
		if (keyText.equals(""))
		{
			return null;
		}

		// Base64 decode it
		byte[] keyBytes = Base64.decode(keyText);

		// Convert the key from a byte array to a PrivateKey

		if (keyEncoding == KEY_ENCODING_PKCS8)
		{
			// Decrypt the key
 			if (keyEncrypted)
			{
				keyBytes = decryptPkcs8PrivateKey(keyBytes);

				// Check if the decryption failed
				if (keyBytes == null)
				{
					return null;
				}
			}

			key = decodePkcs8PrivateKey(keyBytes);
		}
		else if (keyEncoding == KEY_ENCODING_OPENSSL)
		{
			// Decrypt the key
			if (keyEncrypted)
			{
				keyBytes = decryptOpenSSLTraditionalPrivateKey(
					keyBytes, openSSLEncryptionAlgorithm, openSSLSalt);

				// Check if the decryption failed
				if (keyBytes == null)
				{
					return null;
				}
			}

			key = decodeOpenSSLTraditionalPrivateKey(keyBytes, keyAlgo);
		}

		return key;
	}

	/*
	 * Load a PEM encoded X.509 certificate from the supplied stream.
	 *
	 * Returns the certificate if successful, null otherwise.
	 */
	protected X509Certificate loadPEMEncodedCert(InputStream certStream)
	{
		X509Certificate cert = null;

		try
		{
			BufferedInputStream certBIS = new BufferedInputStream(certStream);
			CertificateFactory certCF =
				CertificateFactory.getInstance("X.509");
			cert = (X509Certificate) certCF.generateCertificate(certBIS);
		}
		catch (CertificateException e)
		{
			debug(e);
			System.err.println(
				"Error decoding certificate:  " + e.getMessage());
			return null;
		}

		return cert;
	}

	/*
	 * Attempts to save the specified certificate to the specified
	 * output stream in PEM encoded format.  PEM encoding is a Base64
	 * encoding of DER encoding, with header and footer lines.
	 *
	 * Returns true if successful, false otherwise.
	 */
	protected boolean savePEMEncodedCert(
		Certificate cert,
		OutputStream certStream)
	{
		byte[] certBytes = null;
		try
		{
			// Get the DER encoded cert
			certBytes = cert.getEncoded();
		}
		catch (CertificateEncodingException e)
		{
			debug(e);
			System.err.println(
				"Error encoding certificate:  " + e.getMessage());
			return false;
		}

		// Base64 encode it
		certBytes = Base64.encode(certBytes);
		String certString = new String(certBytes);

		// Wrap the encoded text at 64 characters (It makes it
		// prettier to look at, and OpenSSL won't read it otherwise.)
		certString = characterWrap(certString, 64);

		// Add the header and footer lines
		certString =
			PEM_CERTIFICATE_HEADER + "\n" +
			certString + "\n" +
			PEM_CERTIFICATE_FOOTER;

		// Write the encoded certificate out
		// To ensure portability we specify an encoding instead of
		// letting the system pick one.
		try
		{
			certBytes = certString.getBytes("UTF-8");
			certStream.write(certBytes);
			certStream.close();
		}
		catch (UnsupportedEncodingException e)
		{
			debug(e);
			System.err.println("Error writing cert:  " + e.getMessage());
			return false;
		}
		catch (IOException e)
		{
			debug(e);
			System.err.println("Error writing cert:  " + e.getMessage());
			return false;
		}

		return true;
	}

	/*
	 * Attempts to save the specified key to the specified
	 * output stream in PEM encoded PKCS#8 format.  PEM encoding
	 * is a Base64 encoding of DER encoding, with header and footer
	 * lines.  The specified key must use PKCS#8 as its default
	 * encoding format.
	 *
	 * Returns true if successful, false otherwise.
	 */
	protected boolean savePKCS8EncodedKey(
		PrivateKey privateKey,
		OutputStream keyStream)
	{
		// Make sure the user gave us a key we can work with
		if (! privateKey.getFormat().equals("PKCS8"))
		{
			System.err.println(
				"The primary encoding format of the " +
				"private key is not 'PKCS8', instead " +
				"it is " + privateKey.getFormat() + "." +
				"  We don't support that format.");
			return false;
		}

		// Get the DER encoded key in PKCS#8 format
		byte[] keyBytes = privateKey.getEncoded();

		// Base64 encode it
		keyBytes = Base64.encode(keyBytes);
		String keyString = new String(keyBytes);

		// Wrap the encoded text at 64 characters.  (It makes it
		// prettier to look at, and OpenSSL won't read it otherwise.)
		keyString = characterWrap(keyString, 64);

		// Add the header and footer lines
		keyString =
			PEM_PKCS8_UNENCRYPTED_HEADER + "\n" +
			keyString + "\n" +
			PEM_PKCS8_UNENCRYPTED_FOOTER;

		// Write the encoded key out
		// To ensure portability we specify an encoding instead of
		// letting the system pick one.
		try
		{
			keyBytes = keyString.getBytes("UTF-8");
			keyStream.write(keyBytes);
			keyStream.close();
		}
		catch (UnsupportedEncodingException e)
		{
			debug(e);
			System.err.println("Error writing key:  " + e.getMessage());
			return false;
		}
		catch (IOException e)
		{
			debug(e);
			System.err.println("Error writing key:  " + e.getMessage());
			return false;
		}

		return true;
	}

	/*
	 * Takes a byte array containing a encrypted private key
	 * in PKCS#8 format (EncryptedPrivateKeyInfo) and returns
	 * an unencrypted private key in PKCS#8 format.
	 *
	 * Broken.  There are a couple of possible implementations
	 * commented out here, with comments as to why they don't work.
	 */
	protected byte[] decryptPkcs8PrivateKey(byte[] encodedBytes)
	{
		// Loop until we successfully decrypt the key
		while (true)
		{
			// Ask the user for the passphrase
			char[] passphrase = new char[0];
			while (passphrase.length == 0)
			{
				System.out.print("Passphrase for key:  ");
				passphrase = getInput().toCharArray();
			}

			// *** Until I get this working
			System.err.println(
				"Encrypted PKCS#8 keys are not supported yet");
			return null;

	/*
			// This implementation (using a KeyStore) is a bit of a hack,
			// it was an attempt to get something working until I could
			// get the EncryptedPrivateKeyInfo implementation (below) to
			// work.  Unfortunately, I can't get this to work either.
			// KeyStore doesn't like the already encrypted key, it claims
			// the format is wrong.
			try
			{
				// Create a temp KeyStore
				//KeyStore ks = KeyStore.getInstance("JKS");
				KeyStore ks = KeyStore.getInstance("JCEKS", "SunJCE");
				ks.load(null, null);

				// Insert the key along with a bogus cert just to keep
				// KeyStore happy
				Certificate[] chain =
					new Certificate[]{generatePlaceHolderCert(BigInteger.ONE)};
				ks.setKeyEntry("key", encodedBytes, chain);

				// Now retreive the unencrypted key
				return ((PrivateKey) ks.getKey("key", passphrase)).getEncoded();
			}
			catch(KeyStoreException e)
			{
				debug(e);
				System.err.println("Error decoding key:  " +
					e.getMessage());
				return null;
			}
			catch(NoSuchProviderException e)
			{
				debug(e);
				System.err.println("Error decoding key:  " +
					e.getMessage());
				return null;
			}
			catch(IOException e)
			{
				debug(e);
				System.err.println("Error decoding key:  " +
					e.getMessage());
				return null;
			}
			catch(NoSuchAlgorithmException e)
			{
				debug(e);
				System.err.println("Error decoding key:  " +
					e.getMessage());
				return null;
			}
			catch(CertificateException e)
			{
				debug(e);
				System.err.println("Error decoding key:  " +
					e.getMessage());
				return null;
			}
			catch(UnrecoverableKeyException e)
			{
				debug(e);
				System.err.println("Error decoding key:  " +
					e.getMessage());
				// We don't return for this exception because this
				// is generally an indication that the user gave us
				// the wrong password, so we want to loop and give
				// them another try
				//return null;
			}
	*/

	/*
			// This would be the clean way to do it, but 
			// encKeyInfo.getAlgParameters always returns null, and we
			// need the parameters to decode the key
			try
			{
				EncryptedPrivateKeyInfo encKeyInfo =
					new EncryptedPrivateKeyInfo(keyBytes);

				// Get the encryption algorithm
				String algoOID = encKeyInfo.getAlgName();
				System.out.println("Algo OID is " + algoOID);

				// And the algorithm parameters (the salt
				// and iteration count used in the encryption)
				AlgorithmParameters ap = encKeyInfo.getAlgParameters();
				if (ap == null)
				{
					System.err.println("getAlgParameters() returned null");
					return null;
				}

				// getAlgName() seems to return the OID of the
				// algorithm as defined in PKCS#5, we need to convert
				// to Java's name for the algorithm.
				String algo = mapPkcs5OidToString(algoOID, ap);
				System.out.println("Key is encrypted using:  " + algo);

				// Generate a Passphrase Based Encryption (PBE) key
				// from the user supplied passphrase
				PBEKeySpec ks = new PBEKeySpec(passphrase);
				SecretKeyFactory skf = SecretKeyFactory.getInstance(algo);
				PBEKey pbeKey = (PBEKey) skf.generateSecret(ks);

				// Create a Cipher to decrypt the key with
				Cipher cipher = Cipher.getInstance(algo);

				// Initialize the cipher with the passphrase-based key
				// and algorithm parameters
				cipher.init(Cipher.DECRYPT_MODE, pbeKey, ap);

				pkcs8KeySpec = encKeyInfo.getKeySpec(cipher);
			}
			catch (IOException e)
			{
				debug(e);
				System.err.println("Error decoding key:  " +
					e.getMessage());
				return null;
			}
			catch (NoSuchAlgorithmException e)
			{
				debug(e);
				System.err.println("Error decoding key:  " +
					e.getMessage());
				return null;
			}
			catch (InvalidKeySpecException e)
			{
				debug(e);
				System.err.println("Error decoding key:  " +
					e.getMessage());
				return null;
			}
			catch (NoSuchPaddingException e)
			{
				debug(e);
				System.err.println("Error decoding key:  " +
					e.getMessage());
				return null;
			}
			catch (InvalidKeyException e)
			{
				debug(e);
				System.err.println("Error decoding key:  " +
					e.getMessage());
				return null;
			}
			catch (InvalidAlgorithmParameterException e)
			{
				debug(e);
				System.err.println("Error decoding key:  " +
					e.getMessage());
				return null;
			}
	*/
		}
	}

	/*
	 * Takes a byte array containing a private key encoded in
	 * PKCS#8 format and returns the decoded key.
	 *
	 * Returns the key if successful, null otherwise.
	 */
	protected PrivateKey decodePkcs8PrivateKey(byte[] encodedBytes)
	{
		try
		{
			// ***We shouldn't assume the key is RSA
			// PKCS8EncodedKeySpec should (but doesn't) provide a
			// method to query the type of key.  Maybe we should
			// try DSA in the InvalidKeySpecException catch block?
			PKCS8EncodedKeySpec keySpec =
				new PKCS8EncodedKeySpec(encodedBytes);
			return KeyFactory.getInstance("RSA").generatePrivate(keySpec);
		}
		catch (NoSuchAlgorithmException e)
		{
			debug(e);
			System.err.println(
				"Error decoding key:  " + e.getMessage());
			return null;
		}
		catch (InvalidKeySpecException e)
		{
			debug(e);
			System.err.println(
				"Error decoding key:  " + e.getMessage());
			return null;
		}
	}

	/*
	 * Takes a byte array containing a private key encrypted
	 * in OpenSSL's "traditional" format and returns
	 * the decrypted key in OpenSSL's "traditional" format.
	 * The algorithm should be supplied from the 'DEK-Info' header
	 * line in the PEM encoded file.
	 *
	 * Broken.  As it stands now it compiles and runs, and is coded
	 * as best I can guess it ought to work, but the resulting key
	 * is not valid.  I think perhaps that OpenSSL isn't using the
	 * official PBE algorithm from PKCS#5, but rather their own
	 * way of generating the key from the passphrase, etc.  However,
	 * I've been unable to wrap my head around the OpenSSL code so
	 * I'm not sure what it is they are doing.
	 */
	protected byte[] decryptOpenSSLTraditionalPrivateKey(
		byte[] encryptedBytes, String openSSLAlgorithm, byte[] salt)
	{
		// According to http://www.openssl.org/docs/crypto/pem.html
		// OpenSSL always uses an iteration count of 1 for
		// "traditional" format keys
		final int iterationCount = 1;

		// *** Until I get this working
		System.err.println(
			"Encrypted OpenSSL traditional keys are not supported yet");
		return null;

		/*
		// Loop until we successfully decrypt the key
		while (true)
		{
			// Ask the user for the passphrase
			char[] passphrase = new char[0];
			while (passphrase.length == 0)
			{
				System.out.print("Passphrase for key:  ");
				passphrase = getInput().toCharArray();
			}

			// OpenSSL uses the following encryption algorithms:
			//
			// DES-CBC
			// DES-EDE3-CBC
			// (OpenSSL also supports IDEA, but my copy wasn't compiled
			// with it, so I don't have a sample string for it.)
			//
			// It would be nice if we could automatically map these
			// instead of hard coding the mapping for each one.
			//
			// None of the OpenSSL documentation says what digest
			// algorithm is used (as best I can see), but some digging
			// in the OpenSSL source leads me to believe it is MD5.
			//
			// JCE Reference Guide:
			// http://java.sun.com/j2se/1.4/docs/guide/security/
			//	jce/JCERefGuide.html#AppA
			//
			String jceAlgorithm = null;
			if (openSSLAlgorithm.equals("DES-CBC"))
			{
				jceAlgorithm = "PBEWithMD5AndDES";
			}
			else if (openSSLAlgorithm.equals("DES-EDE3-CBC"))
			{
				jceAlgorithm = "PBEWithMD5AndTripleDES";
			}
			else
			{
				System.err.println(
					"Unrecognized OpenSSL encryption algorithm " +
					openSSLAlgorithm);
				return null;
			}

			try
			{
				// Generate a Passphrase Based Encryption (PBE) key
				// from the user supplied passphrase
				PBEKeySpec keySpec = new PBEKeySpec(
					passphrase, salt, iterationCount);
				SecretKey key =
					SecretKeyFactory.getInstance(jceAlgorithm).
						generateSecret(keySpec);

				// Get the extra parameters ready to feed to the Cipher
				System.out.println("Salt is " +
					salt.length + " bytes: " + new String(salt));
				PBEParameterSpec ps = new PBEParameterSpec(
					salt, iterationCount);
				//IvParameterSpec iv = new IvParameterSpec(salt);

				// Create a Cipher to decrypt the key with
				Cipher cipher = Cipher.getInstance(jceAlgorithm);
				//Cipher cipher = Cipher.getInstance("DES/CBC/NoPadding");
				//Cipher cipher = Cipher.getInstance("DES/CBC/PKCS5Padding");

				// Initialize the cipher with the passphrase-based key
				cipher.init(Cipher.DECRYPT_MODE, key, ps);
				//cipher.init(Cipher.DECRYPT_MODE, key, iv);

				byte[] unencryptedBytes = cipher.doFinal(encryptedBytes);
				System.out.println("Key decrypted");
				// *** Debugging
				try
				{
					FileOutputStream fos = new FileOutputStream("key.decrypt");
					fos.write(unencryptedBytes);
					fos.close();
					System.out.println("Decrypted key written to key.decrypt");
					System.exit(0);
				}
				catch (FileNotFoundException e) {}
				catch (IOException e) {}
				return unencryptedBytes;
			}
			catch (NoSuchAlgorithmException e)
			{
				debug(e);
				System.err.println(
					"Error decrypting key:  " + e.getMessage());
				return null;
			}
			catch (InvalidKeySpecException e)
			{
				debug(e);
				System.err.println(
					"Error decrypting key:  " + e.getMessage());
				return null;
			}
			catch (NoSuchPaddingException e)
			{
				debug(e);
				System.err.println(
					"Error decrypting key:  " + e.getMessage());
				return null;
			}
			catch (InvalidKeyException e)
			{
				debug(e);
				System.err.println(
					"Error decrypting key:  " + e.getMessage());
				return null;
			}
			catch (InvalidAlgorithmParameterException e)
			{
				debug(e);
				System.err.println(
					"Error decrypting key:  " + e.getMessage());
				return null;
			}
			catch (IllegalBlockSizeException e)
			{
				debug(e);
				System.err.println(
					"Error decrypting key:  " + e.getMessage());
				return null;
			}
			catch (BadPaddingException e)
			{
				debug(e);
				System.err.println(
					"Error decrypting key:  " + e.getMessage());
				return null;
			}
		}
		*/
	}

	/*
	 * Takes a byte array containing a private key encoded in
	 * OpenSSL's "traditional" format and returns the decoded key.
	 *
	 * Returns the key if successful, null otherwise.
	 */
	protected PrivateKey decodeOpenSSLTraditionalPrivateKey(
		byte[] encodedBytes, int keyAlgorithm)
	{
		PrivateKey key = null;

		try
		{
			// Use some Bouncy Castle classes to go from the encoded key
			// to an object.

			// Turn the array into a DERInputStream so that we can
			// read DER objects from it
			DERInputStream keyDERStream =
				new DERInputStream(new ByteArrayInputStream(encodedBytes));

			if (keyAlgorithm == KEY_ALGO_RSA)
			{
				// Use BC's RSAPrivateKeyStructure to parse the
				// elements of the key
				RSAPrivateKeyStructure rsaStruct =
					new RSAPrivateKeyStructure(
					(DERConstructedSequence) keyDERStream.readObject());

				// Now jump to standard JCE via a mess
				RSAPrivateCrtKeySpec spec = new RSAPrivateCrtKeySpec(
					rsaStruct.getModulus(),
					rsaStruct.getPublicExponent(),
					rsaStruct.getPrivateExponent(),
					rsaStruct.getPrime1(),
					rsaStruct.getPrime2(),
					rsaStruct.getExponent1(),
					rsaStruct.getExponent2(),
					rsaStruct.getCoefficient());
				key = KeyFactory.getInstance("RSA").generatePrivate(spec);
			}
			else if (keyAlgorithm == KEY_ALGO_DSA)
			{
				// BouncyCastle doesn't have an equivalent to
				// RSAPrivateKeyStructure for DSA so we do it by hand
				DERConstructedSequence seq =
					(DERConstructedSequence) keyDERStream.readObject();
				Enumeration enum = seq.getObjects();
				BigInteger p = ((DERInteger) enum.nextElement()).getValue();
				BigInteger q = ((DERInteger) enum.nextElement()).getValue();
				BigInteger g = ((DERInteger) enum.nextElement()).getValue();
				BigInteger x = ((DERInteger) enum.nextElement()).getValue();

				// Now jump to standard JCE
				DSAPrivateKeySpec spec = new DSAPrivateKeySpec(x, p, q, g);
                   key = KeyFactory.getInstance("DSA").generatePrivate(spec);
			}

			return key;
		}
		catch (IOException e)
		{
			debug(e);
			System.err.println(
				"Error decoding key:  " + e.getMessage());
			return null;
		}
		catch (NoSuchAlgorithmException e)
		{
			debug(e);
			System.err.println(
				"Error decoding key:  " + e.getMessage());
			return null;
		}
		catch (InvalidKeySpecException e)
		{
			debug(e);
			System.err.println(
				"Error decoding key:  " + e.getMessage());
			return null;
		}
	}

	/*
	 ************************************************************************
	 * Misc. methods
	 ************************************************************************
	 */

	/*
	 * Wrap the given string every wrapIndex characters.
	 * I.e. characterWrap("123456789", 3) returns "123\n456\n789"
	 */
	protected String characterWrap(String stringToWrap, int wrapIndex)
	{
		StringBuffer sb = new StringBuffer(stringToWrap);

		for (int i=wrapIndex; i<sb.length(); i=i+wrapIndex+1)
		{
			sb.insert(i, "\n");
		}

		return sb.toString();
	}

	/*
	 * This method takes an org.bouncycastle.asn1.x509.GeneralName
	 * and makes it into a user-friendly string, using a format similar
	 * to that used by OpenSSL (aka the output in the X509v3 Subject
	 * Alternative Name field when you run
	 * 'openssl x509 -text -noout -in file.pem' on a certificate.
	 */
	protected String makeGeneralNamePretty(GeneralName gn)
	{
		int type = ((DERTaggedObject) gn.getDERObject()).getTagNo();
		DERObject obj = ((DERTaggedObject) gn.getDERObject()).getObject();

		String result = "Unrecognized GeneralName";

		// The only two of these I've ever seen via OpenSSL are 1 and 2,
		// so the rest may or may not match what OpenSSL does.
		switch (type)
		{
		case 0:  // OtherName
			// ***
			break;
		case 1:  // rfc822Name
			result = "email:" + ((DERIA5String) obj).getString();
			break;
		case 2:  // dNSName
			result = "DNS:" + ((DERIA5String) obj).getString();
			break;
		case 3:  // x400Address
			// ***
			break;
		case 4:  // directoryName (aka distinguished name)
			// ***
			break;
		case 5:  // ediPartyName
			// ***
			break;
		case 6:  // uniformResourceIdentifier
			result = "URL:" + ((DERIA5String) obj).getString();
			break;
		case 7:  // iPAddress
			try
			{
				InetAddress ip =
					InetAddress.getByAddress(
						((DEROctetString) obj).getOctets());
				result = "IP:" + ip.toString();
			}
			catch (UnknownHostException e)
			{
				System.err.println(
					"makeGeneralNamePretty was passed a GeneralName " +
					"which is supposed to be an IP address but isn't " +
					"the right length");
				result = "error";
			}
			break;
		case 8:  // registeredID
			result = "OID:" + ((DERObjectIdentifier) obj).getId();
			break;
		default:
			// The above cases handle all of the tags defined in
			// RFC 2459, so if we get something else we complain.
			System.err.println(
				"makeGeneralNamePretty was passed a GeneralName with an " +
				"invalid tag");
			result = "error";
			break;
		}

		return result;
	}

	/*
	 * Map from the OIDs defined in sections A.3 and A.4 of PKCS#5 to
	 * Java's name for the algorithm
	 *
	 * I.e mapPkcs5OidToString("1.2.840.113549.1.5.3") returns
	 * "PBEWithMD5AndDES"
	 *
	 * This is incomplete as it doesn't handle PBES2 keys.  For those
	 * the key derivation function and the encryption scheme are
	 * stored within the algorithm parameters.  But since I can't
	 * get EncryptedPrivateKeyInfo to give me the algorithm parameters
	 * for any key (see the comments in the decryptPkcs8PrivateKey
	 * method), I haven't been able to figure out how to retrieve
	 * those.
	 */
	protected String mapPkcs5OidToString(String oid, AlgorithmParameters ap)
	{
		Hashtable oidMap = new Hashtable();
		oidMap.put("1.2.840.113549.1.5.1", "PBEWithMD2AndDES");
		oidMap.put("1.2.840.113549.1.5.3", "PBEWithMD5AndDES");
		oidMap.put("1.2.840.113549.1.5.4", "PBEWithMD2AndRC2");
		oidMap.put("1.2.840.113549.1.5.6", "PBEWithMD5AndRC2");
		oidMap.put("1.2.840.113549.1.5.10", "PBEWithSHA1AndDES");
		oidMap.put("1.2.840.113549.1.5.11", "PBEWithSHA1AndRC2");
		oidMap.put("1.2.840.113549.1.5.13", "PBES2");

		if (oid.equals("1.2.840.113549.1.5.13"))
		{
			System.err.println("AP algo:  " + ap.getAlgorithm());
			System.err.println("AP as string:  " + ap.toString());
			System.err.println("Keys encrypted using the PBES2 scheme defined in v2.0 of PKCS#5 are not currently supported.  I.e. keys generated using the -v2 flag to OpenSSL's pkcs8 command.");
		}

		return (String) oidMap.get(oid);
	}
}

